/* -*- Mode: C -*-
 * Program to do parallel quicksort.  Ben Moszkowski
 * This version takes less virtual time that par-qsort but
 * requires multiple passes.
 * Initial version written around 29 Apr 84.
 * Subsequent changes made.
 * Updated 7 Feb 85.
 */

define sort_test(init_vector) = {
    exists L : {
	fixed_list(L,|init_vector|)
	and L=init_vector
	and par_quicksort(L)
        and monitor(L)
    }
}.

define par_quicksort(L) = {
    if |L| <= 1
    then empty
    else exists pivot : {
	quick_partition(L,pivot);
	sort_parts(L,pivot)
    }
}.

define sort_parts(L,pivot) = {
    exists Ready1,Ready2 : {
	halt (Ready1 and Ready2)
	and quicksort_process(Ready1, L[0 to pivot])
	and quicksort_process(Ready2, L[pivot+1 to |L|])
        }
    and stable L[pivot]
}.

define quicksort_process(Ready,L) =  {
    {
        keep(Ready=false) and
        par_quicksort(L) and fin(Ready=true)
    };
    { 
	 stable L and stable Ready
    }
}.


define monitor(L) = {
    exists Tag_list : {
	fixed_list(Tag_list,|L|) and
	forall i<|L| : {
	    Tag_list[i] is (if i=L[i] then 1 else 0)
	} and
	always format("List=%t  Tag_list=%t\n",L,Tag_list)
    }
}.

define sort_test1 = [2,0,1].

/* run */ define test1() = sort_test(sort_test1).

define sort_test2 = [2,4,9,1,0,10,6,8,7,5,3].

/* run */ define test2() = sort_test(sort_test2).
    
define sort_test3 = [13,7,5,3,1,9,8,6,2,0,12,11,4,10].

/* run */ define test3() = sort_test(sort_test3).
    
define sort_test4 = [4,5,2,0,6,1,3].

/* run */ define test4() = sort_test(sort_test4).    
