/* fns.h:  
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: fns.h,v 2.16 2008/08/13 12:41:40 acau Exp $
 */

/* parser.c */
NODEPTR parse(FILE *stream);
NODEPTR resolve_refs(NODEPTR e);
NODEPTR rexpr(int n);
NODEPTR rbexpr(void);
NODEPTR rexists(void);
NODEPTR rforall(void);
NODEPTR rif(void);
NODEPTR rwhile(void);
NODEPTR rrepeat(void);
NODEPTR rchopstar(void);
NODEPTR rkeep(void);
NODEPTR rfor(void);
NODEPTR rdefine(void);
NODEPTR rset(void);
NODEPTR rlambda(void);
NODEPTR rlist(void);
NODEPTR rsublist(NODEPTR list);
NODEPTR rlistdef(unsigned int type);
NODEPTR rformat(void);
NODEPTR rsave(void);
NODEPTR rhelp(void);
NODEPTR rcall(NODEPTR id);
NODEPTR rnamelist(void);
NODEPTR rname(void);
NODEPTR rexprlist(void);
NODEPTR rsysvar(void);
void push_scope(NODEPTR name_list);
NODEPTR pop_scope(void);
void install_name_list(NODEPTR name_list);
NODEPTR lookup_env(char *s);
NODEPTR install_global_id(char *name);
void install_local_id(char *name);
void next_token(void);
void assert_syntax(int n);
void syntax_error(int n);
void syntax_msg(int n);


/* lexer.c */
int read_token(void);
int isnextchar(int expect);
int follow(int expect, int ifyes, int ifno);
int backslash(int c);
NODEPTR symballoc(void);
unsigned hash(char *s);
NODEPTR lookup(char *s);
NODEPTR install(int type, char *name);
void install_keyword(int type, char *name);


/* init.c */
void install_keywords(void);
void tempura_init (void);


/* top-level.c */
int main(void);
void prompt(int level);
void tempura(int level);
void exec_command(NODEPTR cmd); 
NODEPTR reduce_stmt(NODEPTR e); 
NODEPTR reduce_val(NODEPTR e); 
NODEPTR reduce_ref(NODEPTR e); 
NODEPTR reduce_pgm(NODEPTR e); 
NODEPTR reduce_ref_and_fetch(NODEPTR e); 
NODEPTR make_ref(NODEPTR e);
void get_val(NODEPTR e);
void get_ref(NODEPTR e);


/* version.c */
void print_version_number(void);


/* errors.c */
void pop_tempura_level(int b);
void tempura_prefix(char *s);
void error_break2(char *text, char *v1, char *v2);
void error_break2d(char *text, int v1, int v2);
void error_break2e(char *text, int v1);
void user_break(char *text);
void warn1(char *text, char *v1);
void warn2(char *text, char *v1, char *v2);
void show_state(void);
void show_undef_vars(void);
void state_not_defined(void);
void show_form(void);
void show_trace(void);
void show_scope(void);


/* val.c */
void type_check(NODEPTR e, unsigned type);
void static_check(NODEPTR e);
int val_defined(NODEPTR e);
int pointer_defined(NODEPTR p);
int list_defined(NODEPTR list);
int max_size(NODEPTR list);
int list_size_defined(NODEPTR list);
void output_val(NODEPTR e);
void foutput_val(NODEPTR e);
void output_list(NODEPTR list);
void foutput_list(NODEPTR list);
NODEPTR copy_val(NODEPTR val);
NODEPTR copy_list(NODEPTR list);
int string_size(NODEPTR s);
NODEPTR substring(NODEPTR s, int i, int j);
NODEPTR append_string(NODEPTR s1, NODEPTR s2); 
NODEPTR append_list(NODEPTR list1, NODEPTR list2);
NODEPTR new_list(int cell_type, int size_type, NODEPTR n);
void assign_val (register NODEPTR loc, NODEPTR val);
void assign_struct(NODEPTR loc, NODEPTR struct1);
void assign_list(NODEPTR loc, NODEPTR list);
int equal_val(NODEPTR e1, NODEPTR e2);
int compare_val(NODEPTR e1, NODEPTR e2);
int compare_strings(NODEPTR e1, NODEPTR e2);
int compare_ptrs(NODEPTR e1, NODEPTR e2);
int compare_lists(NODEPTR list1, NODEPTR list2);


/* mem.c */
NODEPTR new_loc(register NODEPTR loc, int attrib);
NODEPTR mem_block(int attrib, register unsigned int size);
NODEPTR allocate_mem(unsigned int size);
int mem_defined(register NODEPTR loc);
NODEPTR get_mem(register NODEPTR loc);
NODEPTR curr_list(NODEPTR list);
NODEPTR next_list(NODEPTR list);
NODEPTR list_size1(NODEPTR list);
NODEPTR list_base1(NODEPTR list);
NODEPTR put_mem(register NODEPTR loc, NODEPTR val);
int is_static_mem(register NODEPTR loc);
void make_static_mem(register NODEPTR loc);
void make_indirect_mem(NODEPTR loc, register NODEPTR val);
void erase_mem(register NODEPTR loc);
int empty_defined(void);
void put_empty(NODEPTR val);
NODEPTR get_empty(void);
void push_env(NODEPTR bnd);
void pop_env(void);
void push_display(NODEPTR bnd);
void pop_display(void);


/* store.c */
char *allocate(unsigned int size, unsigned int elt_size);
char *strsave(char *s);
char *strapnd(char *s1, char *s2);
char *substr(char *s, int i, int j);
void more_symbol_table(unsigned size);
void more_heap(register unsigned size);
NODEPTR claim_node(void);
int free_heap_size(void);
NODEPTR node(unsigned type, NODEPTR v1, NODEPTR v2, NODEPTR v3);
NODEPTR bool_node(unsigned int val);
NODEPTR name_node (unsigned signal, char *name, unsigned int level, unsigned int off);
NODEPTR loc_node (NODEPTR loc, unsigned is_next);
NODEPTR int_node(long val);
NODEPTR string_node (char *val);
NODEPTR env_node (unsigned level, int size, NODEPTR id);
NODEPTR bnd_node (NODEPTR env, NODEPTR blk);
NODEPTR list_node(NODEPTR blk, NODEPTR size_loc, NODEPTR base_loc);
NODEPTR block_node(unsigned int size, NODEPTR base_loc);
unsigned int elist_size(NODEPTR elist);


/* io.c */
void output_int(int n, int f);
void foutput_int(int n, int f);
void output_str(char *s);
void foutput_str(char *s);
void outputf_str(char *s, int f);
void foutputf_str(char *s, int f);
void output_lambda(NODEPTR e);
void foutput_lambda(NODEPTR e);
char *print_name(NODEPTR e);
void message(char *text);
void spaces(int n);
void fspaces(int n);
void output_env(NODEPTR env);
void foutput_env(NODEPTR env);
void output_expn(NODEPTR e, int level);
void foutput_expn(NODEPTR e, int level);
void foutput_char(char c);


/* system.c */
RETSIGTYPE onintr(int sig);
RETSIGTYPE onfpe(int sig);
RETSIGTYPE onbus(int sig);
RETSIGTYPE onsegv(int sig);
RETSIGTYPE onsys(int sig);
RETSIGTYPE onalrm(int sig);
void catch_signals(void);
void breakin(void);
FILE *openfile(char *fullname, char *spath, char *name, char *suffix, char *mode);
int fisspace(int s);
void assign_sysvar(NODEPTR id, NODEPTR val);
void init_sysvars(void);
void start_time(void);
void check_time(void);
int own_wait(int *statusPtr);
void check_sys_waits(void);


/* r-assign.c */
NODEPTR assign_curr(NODEPTR e1, NODEPTR e2);
NODEPTR assign_next(NODEPTR e1, NODEPTR e2);
void reduce_EQ_stmt(NODEPTR e);
void reduce_GETS(NODEPTR e);
void reduce_STB(NODEPTR e);
void reduce_IS(NODEPTR e);
void reduce_ASSN(NODEPTR e);
void reduce_UASSN(NODEPTR e);


/* r-int.c */
void reduce_SEMI(NODEPTR e);
void reduce_iSEMI(NODEPTR e);
void reduce_WHILE(NODEPTR e);
void reduce_REPEAT(NODEPTR e);
void reduce_CHOPSTAR(NODEPTR e);
void reduce_KEEP(NODEPTR e);
void reduce_FORTIMES(NODEPTR e);
void reduce_iFORTIMES(NODEPTR e);
void reduce_FOR(NODEPTR e);
void reduce_iFOR(NODEPTR e);
void reduce_FORIN(NODEPTR e);
void reduce_iFORIN(NODEPTR e);
void reduce_PROJ(NODEPTR e);
void reduce_iPROJ(NODEPTR e);


/* r-linear.c */
void reduce_AND(NODEPTR e);
void reduce_iAND(NODEPTR e);
void reduce_IMPLIES(NODEPTR e);
void reduce_WHERE(NODEPTR e);
void reduce_IF_stmt(NODEPTR e);
void reduce_IF_val(NODEPTR e);
void reduce_iIF_val(NODEPTR e);
void reduce_CALL(NODEPTR e);
void reduce_iCALL_stmt(NODEPTR e);
void reduce_iCALL_val(NODEPTR e);
void reduce_iCALL_ref(NODEPTR e);
void reduce_EXISTS(NODEPTR e);
void reduce_iEXISTS_stmt(NODEPTR e);
void reduce_iEXISTS_val(NODEPTR e);
void reduce_FORALL(NODEPTR e);
void reduce_THE(NODEPTR e);
void reduce_ALWAYS(NODEPTR e);
void reduce_SOMETIME(NODEPTR e);
void reduce_HALT(NODEPTR e);
void reduce_INIT(NODEPTR e);
void reduce_FIN(NODEPTR e);
void reduce_KEEP(NODEPTR e);
void reduce_WNEXT(NODEPTR e);
void reduce_NEXT_stmt(NODEPTR e);
void reduce_NEXT_val(NODEPTR e);
void reduce_NEXT_ref(NODEPTR e);
void reduce_LEN(NODEPTR e);
void reduce_iLEN(NODEPTR e);
void reduce_RANLEN(NODEPTR e);
void reduce_INFINITE(NODEPTR e);
void reduce_PROCESS(NODEPTR e);
void reduce_iPROCESS(NODEPTR e);
void reduce_PREFIX(NODEPTR e);
void reduce_iPREFIX(NODEPTR e);


/* r-symbol.c */
void noop(NODEPTR e);
void assert_stmt(NODEPTR e);
void assert_expr(NODEPTR e);
void assert_stmt_to_true(NODEPTR e);
void assert_true(NODEPTR e);
void reduce_const(NODEPTR e);
void reduce_id_val(NODEPTR e);
void reduce_id_ref(NODEPTR e);
void reduce_EMPTY_stmt(NODEPTR e);
void reduce_MORE_stmt(NODEPTR e);
void reduce_EMPTY_val(NODEPTR e);
void reduce_MORE_val(NODEPTR e);
void reduce_EMPTY_ref(NODEPTR e);
void reduce_SKIP(NODEPTR e);


/* r-expr.c */
void reduce_TYPE(NODEPTR e);
void reduce_UNDEFINED(NODEPTR e);
void reduce_EQ_val(NODEPTR e);
long power(long i, long j);
int umult32(long a,long b,long* c);
int overflow(long a, long b, long *result);
NODEPTR calculate_val(unsigned op, NODEPTR e1, NODEPTR e2);
void reduce_bin_op(NODEPTR e);
void reduce_u_op(NODEPTR e);
void reduce_RANDOM(void);
void reduce_RRANDOM(void);
void reduce_ASCII(NODEPTR e);
void reduce_REFER(NODEPTR e);
void reduce_DEREF(NODEPTR e);


/* r-io.c */
void reduce_INPUT(NODEPTR e);
void reduce_OUTPUT(NODEPTR e);
void reduce_PEEK(NODEPTR e);
void reduce_FORMAT(NODEPTR e);


/* r-lists.c */
void reduce_list(NODEPTR e);
void reduce_STRUCT_val(NODEPTR e);
void reduce_STRUCT_ref(NODEPTR e);
void reduce_MAKELIST(NODEPTR e);
int reduce_index(NODEPTR e);
void reduce_SUB(NODEPTR e);
void reduce_SIZE(NODEPTR e);
NODEPTR sublist(NODEPTR list, int m, int n);
void reduce_SUBLIST(NODEPTR e);


/* r-system.c */
void reduce_RUN(NODEPTR e);
void init_exec_stat(void);
void init_state_stat(void);
void init_pass_stat(void);
void update_exec_stat(void);
void update_state_stat(void);
void output_pass_stat(void);
void output_state_stat(void);
void output_exec_stat(void);
void reduce_LOAD(NODEPTR e);
void reduce_DEFINE(NODEPTR e);
void reduce_EXIT(NODEPTR e);
void reduce_SET(NODEPTR e);
void reduce_SYSVAR(NODEPTR e);
void reduce_SHOWTRACE(NODEPTR e);
void show_mem(NODEPTR loc);
void reduce_SHOWMEM(NODEPTR e);
void reduce_EOF(NODEPTR e);
void reduce_RETURN(NODEPTR e);
void reduce_CONTINUE(NODEPTR e);
void reduce_LEAP(NODEPTR e);
void reduce_STEP(NODEPTR e);
void reduce_BREAK(NODEPTR e);
void reduce_ABORT(NODEPTR e);
void reduce_HELP(NODEPTR e);
void reduce_SYSTEM(NODEPTR e);
void reduce_iSYSTEM(NODEPTR e);
void reduce_SHOWSTATE(NODEPTR e);


/* gc.c */
void mark(NODEPTR e);
void mark_mem(NODEPTR loc, unsigned n);
void mark_phase(void);
int is_refd_block(NODEPTR bp);
void reclaim_node(NODEPTR np);
void reclaim_phase(void);
void collect_garbage(void);




