/* io.c: input and output routines 
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: io.c,v 2.15 2008/08/13 12:41:40 acau Exp $
 */


#include "system.h"



/******************************\
*                              *
*  Auxiliary output routines.  *
*                              *
\******************************/

void output_int(n,f)
int n;
int f;
{
    char    s[MAXINTSIZE],
           *p = s + MAXINTSIZE - 1;
    int     sign;
    *p = '\0';
    if ((sign = n) < 0)
        n = -n;
    do {
        *--p = n % 10 + '0';
        n = n / 10;
    } while (n > 0);
    if (sign < 0)
        *--p = '-';
    outputf_str (p, f);
}

void foutput_int(n,f)
int n;
int f;
{
    char    s[MAXINTSIZE],
           *p = s + MAXINTSIZE - 1;
    int     sign;
    *p = '\0';
    if ((sign = n) < 0)
        n = -n;
    do {
        *--p = n % 10 + '0';
        n = n / 10;
    } while (n > 0);
    if (sign < 0)
        *--p = '-';
    foutputf_str (p, f);
}


void output_str(s)
char *s;
{
    while (*s != '\0')
        output_char (*s++);
}

void foutput_str(s)
char *s;
{
    while (*s != '\0')
        foutput_char (*s++);
}

void outputf_str(s,f)
char *s;
int f;
{
    int     len = strlen (s);
    spaces (f - len);
    output_str (s);
    spaces (-f - len);
}

void foutputf_str(s,f)
char *s;
int f;
{
    int     len = strlen (s);
    fspaces (f - len);
    foutput_str (s);
    fspaces (-f - len);
}

void output_lambda(e)
NODEPTR e;
{
    output_expn(e,-1);
}

void foutput_lambda(e)
NODEPTR e;
{
    foutput_expn(e,-1);
}

/* Find print-name of type. */
char *print_name(e)
NODEPTR e;
{
    unsigned t = node_type(e);
    static char *undef = "undefined";
    return ((0 <= t && t <= MAXTOKEN) ? Keyword[t] : undef);
}

void message(text)
char *text;
{
    printf ("***Tempura: %s.\n", text);
    fflush(stdout);
}

void spaces(n)
int n;
{
    while (n-- > 0)
        output_char (' ');
}

void fspaces(n)
int n;
{
    while (n-- > 0)
        foutput_char (' ');
}

void output_env(env)
NODEPTR env;
{
    NODEPTR id = node_env_idlist(env);
    while (id != NULL) {
        output_str (node_nm_name(id));
        if ((id = node_next(id)) != NULL)
            output_str (", ");
    }
}

void foutput_env(env)
NODEPTR env;
{
    NODEPTR id = node_env_idlist(env);
    while (id != NULL) {
        foutput_str (node_nm_name(id));
        if ((id = node_next(id)) != NULL)
            foutput_str (", ");
    }
}

void output_expn(e,level)
NODEPTR e;
int level;
{

    if (e == NULL) {
/*        output_str ("()");*/
        return;
    }

    if (level == 0) {
        output_str ("...");
        return;
    }

    switch (node_type(e)) {

        case STATE:
        case STATIC:
            output_str (node_nm_name(e));
            break;
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case LIST:
        case LOC:
            output_val (e);
            break;
        case BND:
            output_expn (node_bnd_env(e), level); 
            break;
        case ENV:
            output_env (e); 
            break;
        case MEMBLOCK:
            output_str (Keyword[node_type(e)]);
            output_int ((signed)node_blk_size(e), -1);
            break;
        case SKIP:
        case EMPTY:
        case INFINITE:
        case RANLEN:    
        case MORE:    
            output_str (Keyword[node_type(e)]);
            break;
        case LT:
        case LE:
        case GE:
        case GT:
        case EQ:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case AND:
        case OR:
        case IMPLIES:
        case MOD:
            output_char ('(');
            output_expn (node_lnk_1(e), level);
	    space;
            output_str (Keyword[node_type(e)]);
	    space;
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case SEMI:
        case WHERE:
        case PROJ:
        case GETS:
        case IS:
        case ASSN:
        case UASSN:
            output_expn (node_lnk_1(e), level);
	    space;
            output_str (Keyword[node_type(e)]);
	    space;
            output_expn (node_lnk_2(e), level - 1);
            break;
        case NOT:
        case UPLUS:
        case UMINUS:
        case REFER:
        case DEREF:    
        case PROCESS:
        case iPROCESS:
        case PREFIX:
        case iPREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:
        case NEXT:
        case WNEXT:
        case LEN:
        case iLEN:    
        case STB:
        case INPUT:
        case OUTPUT:
        case RUN:
        case TYPE:
        case ASCII:
        case STRUCT:
        case LOAD:
        case SYSTEM:
            output_str (Keyword[node_type(e)]);
	    space;
            output_expn (node_lnk_1(e), level);
            break;
        case FORMAT:
            output_str (Keyword[node_type(e)]);
	    output_char ('(');;
            output_expn (node_lnk_1(e), level);
	    output_char (')');;
            break;
        case IF:
        case iIF:    
            output_str (Keyword[node_type(e)]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" then {");
            output_expn (node_lnk_2(e), level - 1);
            if (node_lnk_3(e) != NULL) {
                output_str ("} else {");
                output_expn (node_lnk_3(e), level - 1);
            }
            output_char ('}');
            break;
        case CALL:
        case iCALL:
            output_expn (node_lnk_1(e), level);
            output_char ('(');
            output_expn (node_lnk_2(e), level);
            output_char (')');
            break;
        case LAMBDA:
            output_str (Keyword[node_type(e)]);
            output_str (" (");
            output_expn (node_lnk_1(e), level);
            output_str (") : {");
            output_expn (node_lnk_2(e), level - 1);
            output_char ('}');
            break;
        case EXISTS:
        case FORALL:
            output_str (Keyword[node_type(e)]);
            space;
            output_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                output_char ('<');
                output_expn (node_lnk_2(e), level);
            }
            output_str (": {");
            output_expn (node_lnk_3(e), level - 1);
            output_char ('}');
            break;
        case WHILE:
            output_str (Keyword[node_type(e)]);
            output_char (' ');
            output_expn (node_lnk_1(e), level);
            output_str (" do {");
            output_expn (node_lnk_2(e), level - 1);
            output_char ('}');
            break;
        case FOR:
            output_str (Keyword[node_type(e)]);
            space;
            output_expn (node_lnk_1(e), level);
            output_char ('<');
            output_expn (node_lnk_2(e), level);
            output_str (" do {");
            output_expn (node_lnk_3(e), level - 1);
            output_char ('}');
            break;
        case FORIN:
            output_str (Keyword[node_type(e)]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" in ");
            output_expn (node_lnk_2(e), level);
            output_str (" do {");
            output_expn (node_lnk_3(e), level - 1);
            output_char ('}');
            break;
        case FORTIMES:
            output_str ("for ");
            output_expn (node_lnk_1(e), level);
            output_str (" times do {");
            output_expn (node_lnk_2(e), level - 1);
            output_char ('}');
            break;
        case DEFINE:
        case SET:
            output_str (Keyword[node_type(e)]);
            space;
            output_expn (node_lnk_1(e), level);
            output_char ('=');
            output_expn (node_lnk_2(e), level - 1);
            break;
        case COMMA:
	    output_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                output_char (',');
		output_expn (node_lnk_2(e), level);
            }
            break;
        case iEXISTS:
            output_str (Keyword[node_type(e)]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" : {");
            output_expn (node_lnk_2(e), level - 1);
            output_char ('}');
            break;
        case iAND:
        case iSEMI:
        case iPROJ:
            output_expn (node_lnk_2(e), level);
            if (node_lnk_3(e) != NULL) {
                space;
                output_str (Keyword[node_type(e)]);
                space;
                output_expn (node_lnk_3(e), level - 1);
            }
            break;
        case MAKELIST:
            output_char ('[');
            output_expn (node_lnk_1(e), level - 1);
            output_char (']');
            break;
        case SUB:
            output_expn (node_lnk_1(e), level);
            output_char ('[');
            output_expn (node_lnk_2(e), level - 1);
            output_char (']');
            break;
        case SIZE:
            output_char ('|');
            output_expn (node_lnk_1(e), level);
            output_char ('|');
            break;
        case SUBLIST:
            output_expn (node_lnk_1(e), level);
            output_char ('[');
            output_expn (node_lnk_2(e), level - 1);
            output_str ("..");
            output_expn (node_lnk_3(e), level - 1);
            output_char (']');
            break;
        case FIXLIST:
        case VARLIST:
        case iSYSTEM:
            output_str (Keyword[node_type(e)]);
            output_char ('(');
            output_expn (node_lnk_1(e), level - 1);
            output_char (',');
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;    
        default:
            output_str (print_name (e));
            break;
    }
}

void foutput_expn(e,level)
NODEPTR e;
int level;
{

    if (e == NULL) {
/*        output_str ("()");*/
        return;
    }

    if (level == 0) {
        foutput_str ("...");
        return;
    }

    switch (node_type(e)) {

        case STATE:
        case STATIC:
            foutput_str (node_nm_name(e));
            break;
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case LIST:
        case LOC:
            foutput_val (e);
            break;
        case BND:
            foutput_expn (node_bnd_env(e), level); 
            break;
        case ENV:
            foutput_env (e); 
            break;
        case MEMBLOCK:
            foutput_str (Keyword[node_type(e)]);
            foutput_int ((signed)node_blk_size(e), -1);
            break;
        case SKIP:
        case EMPTY:
        case INFINITE:
        case RANLEN:
        case MORE:    
            foutput_str (Keyword[node_type(e)]);
            break;
        case LT:
        case LE:
        case GE:
        case GT:
        case EQ:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case AND:
        case OR:
        case IMPLIES:
        case MOD:
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level);
	    fspace;
            foutput_str (Keyword[node_type(e)]);
	    fspace;
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case SEMI:
        case WHERE:
        case PROJ:
        case GETS:
        case IS:
        case ASSN:
        case UASSN:
            foutput_expn (node_lnk_1(e), level);
	    fspace;
            foutput_str (Keyword[node_type(e)]);
	    fspace;
            foutput_expn (node_lnk_2(e), level - 1);
            break;
        case NOT:
        case UPLUS:
        case UMINUS:
        case REFER:
        case DEREF:    
        case PROCESS:
        case iPROCESS:
        case PREFIX:
        case iPREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:
        case NEXT:
        case WNEXT:
        case LEN:
        case iLEN:
        case STB:
        case INPUT:
        case OUTPUT:
        case RUN:
        case TYPE:
        case ASCII:
        case STRUCT:
        case LOAD:
        case SYSTEM:
            foutput_str (Keyword[node_type(e)]);
	    fspace;
            foutput_expn (node_lnk_1(e), level);
            break;
        case FORMAT:
            foutput_str (Keyword[node_type(e)]);
	    foutput_char ('(');;
            foutput_expn (node_lnk_1(e), level);
	    foutput_char (')');;
            break;
        case IF:
        case iIF:    
            foutput_str (Keyword[node_type(e)]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" then {");
            foutput_expn (node_lnk_2(e), level - 1);
            if (node_lnk_3(e) != NULL) {
                foutput_str ("} else {");
                foutput_expn (node_lnk_3(e), level - 1);
            }
            foutput_char ('}');
            break;
        case CALL:
        case iCALL:
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('(');
            foutput_expn (node_lnk_2(e), level);
            foutput_char (')');
            break;
        case LAMBDA:
            foutput_str (Keyword[node_type(e)]);
            foutput_str (" (");
            foutput_expn (node_lnk_1(e), level);
            foutput_str (") : {");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char ('}');
            break;
        case EXISTS:
        case FORALL:
            foutput_str (Keyword[node_type(e)]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                foutput_char ('<');
                foutput_expn (node_lnk_2(e), level);
            }
            foutput_str (": {");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char ('}');
            break;
        case WHILE:
            foutput_str (Keyword[node_type(e)]);
            foutput_char (' ');
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" do {");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char ('}');
            break;
        case FOR:
            foutput_str (Keyword[node_type(e)]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('<');
            foutput_expn (node_lnk_2(e), level);
            foutput_str (" do {");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char ('}');
            break;
        case FORIN:
            foutput_str (Keyword[node_type(e)]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" in ");
            foutput_expn (node_lnk_2(e), level);
            foutput_str (" do {");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char ('}');
            break;
        case FORTIMES:
            foutput_str ("for ");
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" times do {");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char ('}');
            break;
        case DEFINE:
        case SET:
            foutput_str (Keyword[node_type(e)]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('=');
            foutput_expn (node_lnk_2(e), level - 1);
            break;
        case COMMA:
	    foutput_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                foutput_char (',');
		foutput_expn (node_lnk_2(e), level);
            }
            break;
        case iEXISTS:
            foutput_str (Keyword[node_type(e)]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" : {");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char ('}');
            break;
        case iAND:
        case iSEMI:
        case iPROJ:
            foutput_expn (node_lnk_2(e), level);
            if (node_lnk_3(e) != NULL) {
                fspace;
                foutput_str (Keyword[node_type(e)]);
                fspace;
                foutput_expn (node_lnk_3(e), level - 1);
            }
            break;
        case MAKELIST:
            foutput_char ('[');
            foutput_expn (node_lnk_1(e), level - 1);
            foutput_char (']');
            break;
        case SUB:
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('[');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (']');
            break;
        case SIZE:
            foutput_char ('|');
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('|');
            break;
        case SUBLIST:
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('[');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_str ("..");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (']');
            break;
        case FIXLIST:
        case VARLIST:
        case iSYSTEM:
            foutput_str (Keyword[node_type(e)]);
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level - 1);
            foutput_char (',');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;    
        default:
            foutput_str (print_name (e));
            break;
    }
}

void foutput_char (c)
char c;
{
    char   name[MAXFILENAMESIZE];
    FILE * fp = NULL,
         *openfile ();
    if (strcmp (node_strval(OutFile), "stdout") == 0) {
        output_char (c);
    }
    else {
        if ((fp = openfile (name, node_strval(Path), node_strval(OutFile), FXOUT, "a")) == NULL) {
            warn1 ("cannot open file \"%s\"",  node_strval(OutFile));
        warn0 ("will display to stdout");
        output_char (c);
        }
        else {
           putc(c,fp);
           CharsOutput++;
           fclose (fp);
        }
    }
}
