/* mem.c: Memory management routines 
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: mem.c,v 2.11 2008/08/13 12:41:40 acau Exp $
 */


#include "system.h"





/*******************************\
*                               *
*  Memory management routines   *
*                               *
\*******************************/

/* new_loc:  initialise a memory cell. */
NODEPTR new_loc (loc,attrib)
register NODEPTR loc;
int attrib;
{
    register NODEPTR loc1;
    node_mem_memno(loc) = 0;
    node_mem_memno(loc1 = loc + 1) = 1;
    node_mark(loc) = node_mark(loc1) = F;
    node_mem_isstatic(loc) = node_mem_isstatic(loc1) = (attrib == STATIC);
    node_mem_isindirect(loc) = node_mem_isindirect(loc1) = F;
    node_mem_time(loc) = node_mem_time(loc1) = UNDEF;
    node_next(loc) = node_next(loc1) = NULL;
    return (loc);
}

NODEPTR mem_block(attrib,size)
int attrib;
register unsigned int size;
{
    NODEPTR blk;
    register NODEPTR loc;

    blk = allocate_mem (size);
    loc = node_blk_base(blk);
    while (size-- > 0) {
        new_loc (loc, attrib);
        loc += offset (1);
    }
    return (blk);
}

NODEPTR allocate_mem(size)
unsigned int size;
{
    NODEPTR base_loc;

    if (size == 0)
        return (block_node (0, NULL));
    base_loc = nodep(allocate (offset (size), sizeof (NODE)));
    LocalMemUsed += size;
    return (block_node (size, base_loc));
}

int mem_defined(loc)
register NODEPTR loc;
{
    dereference (loc);
    return (node_mem_isstatic(loc) ?
            (node_mem_time(loc) != UNDEF) :
            (node_mem_time(loc) == (int)(is_next (loc) ? Clock + 1 : Clock)));
}

NODEPTR get_mem (loc)
register NODEPTR loc;
{
    dereference (loc);
    return (node_next(loc));
}

NODEPTR curr_list(list)
NODEPTR list;
{
    return ((is_next (node_lst_size(list)) || is_next (node_lst_base(list))) ?
            list_node (node_lst_blk(list),
                        curr_loc (list_size (list)),
                        curr_loc (list_base (list))) :
             list);
}

NODEPTR next_list(list)
NODEPTR list;
{
    return ((is_curr (node_lst_size(list)) || is_curr (node_lst_base(list))) ?
            list_node (node_lst_blk(list),
                        next_loc (list_size (list)),
                        next_loc (list_base (list))) :
             list);
}

NODEPTR list_size1(list)
NODEPTR list;
{
    NODEPTR loc = node_lst_size(list);
    dereference (loc);
    return (loc);
}

NODEPTR list_base1(list)
NODEPTR list;
{
    NODEPTR loc = node_lst_base(list);
    dereference (loc);
    return (loc);
}


NODEPTR put_mem (loc,val)
register NODEPTR loc;
NODEPTR val;
{
    NODEPTR adj_val;
    dereference (loc);
    if (is_next (loc)) {
        node_next(loc) = adj_val = next_val (val);
        node_mem_time(loc) = Clock + 1;
        if (node_mem_isstatic(loc)) {
            node_next(loc = curr_loc (loc)) = curr_val (val);
            node_mem_time(loc) = Clock;
	}
    }
    else {
        node_next(loc) = adj_val = curr_val (val);
        node_mem_time(loc) = Clock;
	if (node_mem_isstatic(loc)) {
	    node_next(loc = next_loc (loc)) = next_val (val);
            node_mem_time(loc) = Clock + 1;
	}
    }
    return (adj_val);
}

int is_static_mem(loc)
register NODEPTR loc;
{
    dereference (loc);
    return (node_mem_isstatic(loc));
}

void make_static_mem(loc)
register NODEPTR loc;
{
    dereference (loc);
    node_mem_isstatic(next_loc (loc))= node_mem_isstatic(curr_loc (loc)) = T;
}

void make_indirect_mem(loc,val)
NODEPTR loc;
register NODEPTR val;
{
    NODEPTR cloc, nloc;
    dereference (val);
    cloc = curr_loc (loc);
    nloc = next_loc (loc);
    node_next(cloc) = val;
    node_next(nloc) = is_curr (val) ? next_loc (val) : curr_loc (val);
    node_mem_isindirect(cloc) = node_mem_isindirect(nloc) = T;
}

void erase_mem(loc)
register NODEPTR loc;
{
    dereference(loc);
    node_mem_time(loc) = UNDEF;
    node_next(loc) = NULL;
}



/******************************\
*                              *
*  Routines for the empty-flag *
*                              *
\******************************/

int empty_defined()
{
    return (mem_defined (EmptyLoc));
}

void put_empty(val)
NODEPTR val;
{
    if (mem_defined (EmptyLoc)) {
        assert0 (equal_val (val, get_mem (EmptyLoc)),
                "attempt to re-assign interval length");
    }
    else
        put_mem (EmptyLoc, val);
}

NODEPTR get_empty()
{
    return (node_next(EmptyLoc));
}



/*************************\
*                         *
*  Environment handling   *
*                         *
\*************************/

void push_env(bnd)
NODEPTR bnd;
{
    if (EnvTop >= MAXENVSTACKSIZE)
        error_break0("environment stack full");
    else {
        EnvTop += 1;
        EnvStack[EnvTop].env = node_bnd_env(bnd);
        EnvStack[EnvTop].env_base = node_blk_base(node_bnd_blk(bnd));
    }
}

void pop_env()
{
    EnvTop -= 1;
}

void push_display(bnd)
NODEPTR bnd;
{
    register unsigned l = node_env_level(node_bnd_env(bnd));
    unsigned newenvbase;
    if (EnvTop + l + 1 >= MAXENVSTACKSIZE)
        error_break0 ("environment stack full");
    else {
	newenvbase = (EnvTop += 1);
        for (; l > 0; l -= 1) {
            EnvStack[EnvTop].env = EnvStack[EnvBase].env;
            EnvStack[EnvTop].env_base = EnvStack[EnvBase].env_base;
            EnvTop += 1;
            EnvBase += 1;
	}
        EnvStack[EnvTop].env_base = node_blk_base(node_bnd_blk(bnd));
        EnvStack[EnvTop].env = node_bnd_env(bnd);
        EnvBase = newenvbase;
    }
}

void pop_display()
{
    EnvTop = --EnvBase;
    EnvBase = EnvBase - (node_env_level(EnvStack[EnvTop].env));
}


