/* r-expr.c:  expression reduction routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-expr.c,v 2.12 2008/08/13 12:41:40 acau Exp $
 */


#include "system.h"




/****************************************\
*                                        *
*  statements and expressions for types  *
*                                        *
\****************************************/


/* Get the type of something. */
/* type e */
void reduce_TYPE(e)
NODEPTR e;
{
    NODEPTR loc =reduce_ref  (node_lnk_1(e));

    if (Ready) {
       if ((Ready = mem_defined (loc)))
           result (string_node (strsave (print_name (get_mem (loc)))));
    }
}


/* Check whether e is undefined. */
/* undefined e */
void reduce_UNDEFINED(e)
NODEPTR e;
{
     NODEPTR loc=reduce_ref (node_lnk_1(e));

     if ((Ready && mem_defined (loc)) || ForceOutput) {
         if (Ready) {
             if (mem_defined (loc))
                 result(False);
             else
                 result(True);
         }
     }
     
    
    Ready = (Ready &&  mem_defined (loc)) || ForceOutput;
}

/**************************************\
*                                      *
*  Simple Boolean, arithmetic, string  *
*  and list expressions.               *
*                                      *
\**************************************/

/*Equality and Non-equality. */
/* e1 = e2, e1 ~= e2 */
void reduce_EQ_val(e)
NODEPTR e;
{
    NODEPTR e1, e2;
    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        e2 = reduce_val (node_lnk_2(e));
        if (Ready) {
            if (node_type(e) == EQ)
                result (equal_val (e1, e2) ? True : False);
            else
                result (equal_val (e1, e2) ? False : True);
        }
    }
}




#define __HALF_MAX_SIGNED(type) ((type)1 << (sizeof(type)*8-2))
#define __MAX_SIGNED(type) (__HALF_MAX_SIGNED(type) - 1 + __HALF_MAX_SIGNED(type))
#define __MIN_SIGNED(type) (-1 - __MAX_SIGNED(type))

#define __MIN(type) ((type)-1 < 1?__MIN_SIGNED(type):(type)0)
#define __MAX(type) ((type)~__MIN(type))

#define assigncheck(dest,src) ({ \
  typeof(src) __x=(src); \
  typeof(dest) __y=__x; \
  (__x==__y && ((__x<1) == (__y<1))?(void)((dest)=__y),0:1); \
})

#define add_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MIN(typeof(c))-(__b)<=(__a))?assigncheck(c,__a+__b):1) : \
    ((__MAX(typeof(c))-(__b)>=(__a))?assigncheck(c,__a+__b):1); \
})

#define sub_of(c,a,b) ({ \
  typeof(a) __a=a; \
  typeof(b) __b=b; \
  (__b)<1 ? \
    ((__MAX(typeof(c))-(__b)>=(__a))?assigncheck(c,__a-__b):1) : \
    ((__MIN(typeof(c))+(__b)<=(__a))?assigncheck(c,__a-__b):1); \
})

int mult32(long a,long b,long* c) {
    long long x=(long long)a*b;
    if (x!=(long)x) return 0;
  *c=x;
  return 1;
}

long power(i,j)
long i,j;
{
    long p = 1;
    long product;
    while (j-- > 0){
        assert0(mult32 (p, i, &product) == 1,"integer overflow in power" );
        p = product;
    }
    return (p);
}

#define int_plus(X,Y) int_node (node_ival(X) + node_ival(Y))
#define int_minus(X,Y) int_node (node_ival(X) - node_ival(Y))
#define int_mult(X,Y) int_node (node_ival(X) * node_ival(Y))
#define int_div(X,Y) int_node (node_ival(X) / node_ival(Y))
#define int_mod(X,Y) int_node (mod (node_ival(X), node_ival(Y)))
#define int_exp(X,Y) int_node (power (node_ival(X), node_ival(Y)))

/*  calculate value */
/* e1 and e2, e1 or e2 */
/* e1 < e2, e1 <= e2, e1 >= e2, e1 > e2 */ 
/* e1 + e2, e1 - e2, e1 * e2, e1 div e2, e1 mod e2, e1 ** e2 */
NODEPTR calculate_val(op, e1, e2)
unsigned op;
NODEPTR e1, e2;
{
    unsigned type = node_type(e1);
    long val;
    long product;

    
      type_check (e2, type);
      switch (op) {
      case AND:
	switch (type) {
	case BOOL:
            return (node_bval(e1) ? e2 : e1);
	default:
	  break;
	}
	break;
      case OR:
	switch (type) {
	case BOOL:
            return (node_bval(e1) ? e1 : e2);
	default:
	  break;
	}
	break;
      case LT:
          return (compare_val (e1, e2) < 0 ? True : False);
      case LE:
          return (compare_val (e1, e2) <= 0 ? True : False);
      case GE:
          return (compare_val (e1, e2) >= 0 ? True : False);
      case GT:
          return (compare_val (e1, e2) > 0 ? True : False);
      case PLUS:
	switch (type) {
	case INT:
            assert0(add_of(val,node_ival(e1),node_ival(e2))==0,"integer overflow in addition");
	  return (int_node (val));
	case LIST:
            return (append_list (e1, e2));
	case STRING:
            return (append_string (e1, e2));
	default:
	  break;
	}
	break;
      case MINUS:
	switch (type) {
	case INT:
            assert0(sub_of(val,node_ival(e1),node_ival(e2))==0,"integer overflow in subtraction");
	  return (int_node (val));
	default:
	  break;
	}
	break;
      case MULT:
	switch (type) {
	case INT:
            assert0(mult32 (node_ival(e1), node_ival(e2), &product) == 1,"integer overflow in multiplication" );
            return (int_node(product));
	default:
	  break;
	}
	break;
      case DIV:
	switch (type) {
	case INT:
            assert0(node_ival(e2) !=0, "div by zero is undefined\n"); 
	  return (int_div (e1, e2));
	default:
	  break;
	}
	break;
      case MOD:
	switch (type) {
	case INT:
	  return (int_mod (e1, e2));
	default:
	  break;
	}
	break;
      case POWER:
	switch (type) {
	case INT:
            assert0(node_ival(e2) >=0, "power to should be >=0\n");
            product=power(node_ival(e1), node_ival(e2));
	  return (int_node (product));
	default:
	  break;
	}
	break;
      default:
	break;
      }
      error_break2 ("\"%s\" is not a %s operator",
		    Keyword[op], Keyword[type]);
      return (False);
}


/* Binary expressions: AND, OR, PLUS, MINUS, ... */
/* boolean expressions are speeded up otherwise use calculate_val*/
void reduce_bin_op(e)
NODEPTR e;
{
    NODEPTR calculate_val ();
    NODEPTR e1, e2;
    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        switch (node_type(e)) {
            case AND:
                e2 = (node_bval(e1) == F) ? e1 : reduce_val (node_lnk_2(e));
                break;
            case OR:
                e2 = (node_bval(e1) == T) ? e1 : reduce_val (node_lnk_2(e));
                break;
            default:
                e2 = reduce_val (node_lnk_2(e));
        }
    }
    if (Ready)
        result (calculate_val (node_type(e), e1, e2));
}

#define int_uminus(X) int_node (- node_ival(X))
#define int_not(X) (node_ival(X) == 0 ? One : Zero)

/* unary operators: NOT, UPLUS, UMINUS, SIZE, ... */
/* not e1, +e1, -e1, |e1| */
void reduce_u_op(e)
NODEPTR e;
{
    NODEPTR e1;

    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        switch (node_type(e1)) {
            case BOOL:
                switch (node_type(e)) {
                    case NOT:
                        result (node_bval(e1) ? False : True);
                        break;
                    default:
                        error_break1 ("\"%s\" is not a Boolean operator",
                                print_name (e));
                }
                break;
            case INT:
                switch (node_type(e)) {
                    case UPLUS:
                        result (e1);
                        break;
                    case UMINUS:
                        result (int_uminus (e1));
                        break;
                    default:
                        error_break1 ("\"%s\" is not an integer operator",
                                   print_name (e));
                }
                break;
            case STRING:
                switch (node_type(e)) {
                    case SIZE:
                        result (int_node ((int)node_lnk_2(e)));
                        break;
                    default:
                        error_break1 ("\"%s\" is not a string operator",
                                print_name (e));
                }
                break;
            default:
                error_break2 ("\"%s\" is not a %s operator",
                        print_name (e), print_name (e1));
        }
    }
}

/* random  */
/* Reduces to random integer >=0  AC 29Apr1999 */
void reduce_RANDOM()
{
  if ( RandomValTime != Clock || RandomVal == -1 ) {
    RandomVal = rand ();
    RandomValTime = Clock;
  };
  Ready = T;
  result (int_node (RandomVal));
}

/* Random  */
/* Reduces to random integer >=0  AC 20/09/2000 */
/* This will  allow for more than 1 random number
   in the same state */
void reduce_RRANDOM()
{
  if ( RRandomValTime != Clock || RRandomVal == -1 ) {
    RRandomValTime = Clock;
  };
  Ready = T;
  result (int_node (rand()));
}


/********************\
*                    *
*  string operators  *
*                    *
\********************/

/* ascii(e1) */
void reduce_ASCII(e)
NODEPTR e;
{
    char   *strr;
    NODEPTR e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (e1, STRING);
        assert1 (strlen (strr = node_strval(e1)) == 1,
                "cannot get ascii code of string \"%s\" (length not 1)", strr);
        result (int_node (*strr));
    }
}



/**********************\
*                      *
*  pointer operations  *
*                      *
\**********************/
/* ref(loc) */
void reduce_REFER(e)
NODEPTR e;
{
    NODEPTR loc = reduce_ref (node_lnk_1(e));
    if (Ready) {
        assert0 (loc != TempLoc, "cannot reference a value");
        result (loc_node (loc, is_next(loc))); 
    }
}
/* deref p */
void reduce_DEREF(e)
NODEPTR e;
{
    NODEPTR p = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (p, LOC); 
        ref_result (node_loc_tonext(p) ? next_loc (node_loc_ptr(p)) : curr_loc (node_loc_ptr(p)));
    }
}

