/* r-symbol.c: symbol reduction routines
 * Copyright (C) 1987  Roger Hale
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-symbol.c,v 2.11 2008/08/13 12:41:39 acau Exp $
 */


#include "system.h"




void noop(e)
NODEPTR e;
{
    error_break1 ("I'm sorry, Dave, I can't do \"%s\"", print_name (e));
}

void assert_stmt(e)
NODEPTR e;
{
    error_break1 ("\"%s\" can only be executed as a statement",print_name(e));
}

void assert_expr(e)
NODEPTR e;
{
    error_break1 ("\"%s\" cannot be executed as a statement", print_name(e));
    result (True);
}

/* 06/09/2001 added */
void assert_stmt_to_true(e)
NODEPTR e;
{
    NODEPTR e1;
    e1 = reduce_val (e);
    if (Ready) {
	if (e1 != True)
	  error_break0 ("Statement has reduced to false");
	e = True;
    };
    result(e);
}

void assert_true(e)
NODEPTR e;
{
    NODEPTR e1;
    e = reduce_val (e);
    if (!Ready) {
        switch (node_type(e)) {
            case NOT:
	        switch (node_type(e1 = node_lnk_1(e)) ) {
		    case OR:
	                Retry = T;
		        e = node2 (AND,
		                   node1 (NOT, node_lnk_1(e1)),
			           node1 (NOT, node_lnk_2(e1)));
                        break;
		    case AND:
	                Retry = T;
		        e = node2 (OR,
		                   node1 (NOT, node_lnk_1(e1)),
			           node1 (NOT, node_lnk_2(e1)));
                        break;
		    case NOT:
	                Retry = T;
		        e = node_lnk_1(e1);
		        break;
                    default:
		        e1 = assign_curr (node_lnk_1(e), False);
			if (Ready)
			    e = e1;
	                break;
                }
	        break;
	    case OR:
	        e1 = reduce_val (node_lnk_1(e));
	        if (Ready) {
		    if (e1 == True)
		        e = True;
		    else {
		        Retry = T;
		        e = node_lnk_2(e);
		    }
	        } else {
		    e1 = reduce_val (node_lnk_2(e));
	            if (Ready) {
		        if (e1 == True)
		            e = True;
		        else {
		            Retry = T;
		            e = node_lnk_1(e);
		        }
		    }
		}
	        break;
	    default:
	        e1 = assign_curr (e, True);
		if (Ready)
		    e = e1;
	        break;
        }
    }
    result (e);
}


/*******************\
*                   *
*  tempura symbols  *
*                   *
\*******************/

void reduce_const(e)
NODEPTR e;
{
    Ready = val_defined (e);
}

void reduce_id_val(e)
NODEPTR e;
{
    NODEPTR loc;
    NODEPTR val;
    if ((Ready = (mem_defined (loc = adj_loc (mem_loc (e))) &&
                  val_defined (val = get_mem (loc))))) 
       result (val);
}

void reduce_id_ref(e)
NODEPTR e;
{   
    Ready = T;
    ref_result (adj_loc (mem_loc (e)));
}

void reduce_EMPTY_stmt(e)
NODEPTR e;
{
    put_empty (True);
    Ready = T;
    result (True);
}

void reduce_EMPTY_val(e)
NODEPTR e;
{
    NODEPTR empty_loc = adj_loc (EmptyLoc);
    if ((Ready = mem_defined (empty_loc)))
        result (get_mem (empty_loc));
}

void reduce_EMPTY_ref(e)
NODEPTR e;
{
    Ready = T;
    ref_result (adj_loc (EmptyLoc));
}

void reduce_MORE_stmt(e)
NODEPTR e;
{
    put_empty (False);
    Ready = T;
    result (True);
}

void reduce_MORE_val(e)
NODEPTR e;
{
    NODEPTR e1;

    NODEPTR empty_loc = adj_loc (EmptyLoc);
    if ((Ready = mem_defined (empty_loc))) {
      e1 = get_mem (empty_loc);
      result (node_bval(e1) ? False : True);
    }
}

/* unit interval */
void reduce_SKIP(e)
NODEPTR e;
{
    put_empty (False);
    Ready = T;
    result (Empty);
}
